
DeclareRepresentation(
 "IsForestAlgebraRep",
 IsComponentObjectRep and IsAttributeStoringRep,
 ["horizontal","vertical","action"]
);

InstallGlobalFunction(
 ForestAlgebra,
 function(Horizontal, Vertical, Action)
  local FA, alg, dim, gen, n, i, j;
  
  if not IsSemigroup(Horizontal) then
   Error("\nFirst argument must be a valid horizontal semigroup.\n");
  elif not IsSemigroup(Vertical) then
   Error("\nSecond argument must be a valid vertical semigroup.\n");
  elif not IsMatrix(Action) then
   Error("\nThe third argument must be a valid action matrix.\n");
  fi;

  gen := Length(GeneratorsOfSemigroup(Vertical));
  dim := DimensionsMat(Action);
  n := Size(Horizontal);
  if not dim[1]=gen then
   Error("\nThe number of rows of the action matrix should be equal to the number of generators of V.\n");
  fi;

  for i in Action do
   for j in i do
    if not j in [1..n] then
     Error(Concatenation("\nUnexpected ", String(j), " on line ", String(i)),"\n");
    fi;
   od;
  od;

  FA:= NewFamily("ForestAlgebra", IsForestAlgebra);
  alg := rec(
   horizontal := ShallowCopy(Horizontal),
   vertical := ShallowCopy(Vertical),
   action := ShallowCopy(Action)
  );

  return Objectify(
   NewType(FA, IsForestAlgebra and IsForestAlgebraRep and IsAttributeStoringRep),
   alg
  );   
 end
);

InstallGlobalFunction(
 ForestHorizontalMonoid,
 function(HV)
  return HV!.horizontal;
 end
);

InstallGlobalFunction(
 ForestVerticalMonoid,
 function(HV)
  return HV!.vertical;
 end
);

InstallGlobalFunction(
 ForestAction,
 function(HV)
  return HV!.action;
 end
);

InstallGlobalFunction(
 ForestApplyAction,
 function(HV, e1, e2)
  local h, v, H, V, i, G, A;
  H := HV!.horizontal;
  V := HV!.vertical;
  A := HV!.action;
  G := GeneratorsOfSemigroup(V);

  if IsPosInt(e1) and e1 < Size(H) then h:=e1;
  elif e1 in H then h:=Position(Elements(H), e1);
  else Error("Second argument should be an element index or element from H."); fi;

  if not e2 in V then Error("Third argument should be an element from V.");
  else v:=SemigroupFactorization(V, e2)[1]; fi;

  for i in Reversed(v) do
   h := A[Position(G, i)][h];
  od;

  return Elements(H)[h];
 end
);

InstallMethod(ViewObj,
 "displays a forest algebra",
 true, [IsForestAlgebra and IsForestAlgebraRep], 0,
 function(HV)
  Print(
   Concatenation(
    "<H has ", String(Length(GeneratorsOfSemigroup(HV!.horizontal))), " generators and ",
    String(Size(HV!.horizontal)), " elements,",
    " V has ", String(Length(GeneratorsOfSemigroup(HV!.vertical))), " generators>"
   )
  );
 end
);

InstallMethod(Display,
 "displays the multiplication tables of forest algebras",
 true, [IsForestAlgebra and IsForestAlgebraRep], 0,
 function(HV)
  Display(MultiplicationTable(HV!.horizontal));
  Display(MultiplicationTable(HV!.vertical));
  Display(HV!.action);
 end
);

InstallGlobalFunction(
 IsHCommutative,
 function(HV)
  return IsCommutative(HV!.horizontal);
 end
);

InstallGlobalFunction(
 IsVCommutative,
 function(HV)
  return IsCommutative(HV!.vertical);
 end
);

InstallGlobalFunction(
 IsHAperiodic,
 function(HV)
  return Size(HV!.horizontal)=Length(GreensHClasses(HV!.horizontal));
 end
);

InstallGlobalFunction(
 IsVAperiodic,
 function(HV)
  return Size(HV!.vertical)=Length(GreensHClasses(HV!.vertical));
 end
);

InstallGlobalFunction(
 IsHJTrivial,
 function(HV)
  return Size(HV!.horizontal)=Length(GreensJClasses(HV!.horizontal));
 end
);

InstallGlobalFunction(
 IsVJTrivial,
 function(HV)
  return Size(HV!.vertical)=Length(GreensJClasses(HV!.vertical));
 end
);

InstallGlobalFunction(
 IsVRTrivial,
 function(HV)
  return Size(HV!.vertical)=Length(GreensRClasses(HV!.vertical));
 end
);

InstallGlobalFunction(
 IsVLTrivial,
 function(HV)
  return Size(HV!.vertical)=Length(GreensLClasses(HV!.vertical));
 end
);

InstallGlobalFunction(
 IsHRTrivial,
 function(HV)
  return Size(HV!.horizontal)=Length(GreensRClasses(HV!.horizontal));
 end
);

InstallGlobalFunction(
 IsHLTrivial,
 function(HV)
  return Size(HV!.horizontal)=Length(GreensLClasses(HV!.horizontal));
 end
);

InstallGlobalFunction(
 IsHIdempotent,
 function(HV)
  return Size(HV!.horizontal)=Length(Idempotents(HV!.horizontal));
 end
);

InstallGlobalFunction(
 IsVIdempotent,
 function(HV)
  return Size(HV!.vertical)=Length(Idempotents(HV!.vertical));
 end
);